#
#    Copyright 2020, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import os
from json.decoder import JSONDecodeError
from typing import Optional

import requests
from eth_typing import ChecksumAddress
from web3 import Web3

from metemcyber.cli.constants import APP_DIR
from metemcyber.core.bc.account import Account
from metemcyber.core.logger import get_logger
from metemcyber.core.solver import BaseSolver
from metemcyber.core.util import merge_config

LOGGER = get_logger(name='gcs_solver', file_prefix='core')

CONFIG_SECTION = 'gcs_solver'
DEFAULT_CONFIGS = {
    CONFIG_SECTION: {
        'assets_path': f'{APP_DIR}/workspace/upload',
        'functions_url': 'https://exchange.metemcyber.ntt.com',
        'functions_token': 'YOUR_TOKEN_TO_UPLOAD_GCS',
    }
}
# Note: assets_path should be same with "{general.workspace}/upload".


class Solver(BaseSolver):
    def __init__(self, account: Account, operator_address: ChecksumAddress,
                 config_file: Optional[str]) -> None:
        super().__init__(account, operator_address)
        self.config = merge_config(config_file, DEFAULT_CONFIGS, self.config)
        try:
            url = self.config[CONFIG_SECTION]['functions_url']
            token = self.config[CONFIG_SECTION]['functions_token']
            assert url and token not in (
                None, '', DEFAULT_CONFIGS[CONFIG_SECTION]['functions_token'])
        except Exception as err:
            raise Exception('Not enough configuration to upload to GCS') from err
        self.uploader = Uploader(url, token)

    def notify_first_accept(self):
        url = self.config[CONFIG_SECTION]['functions_url']
        return f'Caution: solved challenge data will be uploaded onto {url}.'

    def process_challenge(self, token_address, event):
        LOGGER.info('GCSSolver: callback: %s', token_address)
        LOGGER.debug(event)

        task_id = event['args']['taskId']
        challenge_seeker = event['args']['from']
        LOGGER.info(
            'accepting task %s from seeker %s', task_id, challenge_seeker)
        if not self.accept_task(task_id):
            LOGGER.warning('could not accept task %s', task_id)
            return

        LOGGER.info('accepted task %s', task_id)
        data = ''
        try:
            try:
                download_url = self.upload_to_storage(token_address)
                webhook_url = Web3.toText(event['args']['data'])
            except Exception:
                data = 'Challenge failed by solver side error'
                raise
            try:
                # return answer via webhook
                self.webhook(webhook_url, download_url, challenge_seeker, task_id, token_address)
            except Exception as err:
                data = 'cannot sendback result via webhook: ' + str(err)
                raise
        except Exception as err:
            LOGGER.exception(err)
            LOGGER.error('failed task %s', task_id)
        finally:
            self.finish_task(task_id, data)
            LOGGER.info('finished task %s', task_id)

    def upload_to_storage(self, cti_address):
        file_path = os.path.abspath('{}/{}'.format(
            self.config.get(CONFIG_SECTION, 'assets_path'), cti_address))
        url = self.uploader.upload_file(file_path)
        return url


class Uploader:
    def __init__(self, url: str, functions_token: str) -> None:
        assert url
        assert functions_token
        self.url = url
        self.functions_token = functions_token

    def upload_file(self, upload_path: str) -> str:
        headers = {
            'Authorization': 'Bearer {}'.format(self.functions_token),
            'Content-Type': 'application/json'}
        try:
            # Note: gcs(exchange.metemcyber.ntt.com) accepts json data only.
            with open(upload_path, 'r') as fin:
                jdata = json.load(fin)
        except Exception as err:
            raise Exception(f'Not a expected data format: {upload_path}') from err
        response = requests.post(
            self.url,
            json=jdata,
            headers=headers)
        try:
            results = response.json()
        except JSONDecodeError as err:
            LOGGER.debug(response.headers)
            LOGGER.debug(response.text[:256])
            raise Exception(f'Received unexpected (not a JSON) result: {err}') from err
        if 'result' in results:
            return results['result']  # OK
        raise Exception('File upload failed: ' + results['error'])



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import argparse
from typing import List, Tuple

from metemcyber.core.bc.account import Account
from metemcyber.core.bc.ether import Ether
from metemcyber.core.bc.util import decode_keyfile
from metemcyber.core.logger import get_logger
from metemcyber.core.multi_solver import MCSServer, SolverManager, mcs_console

LOGGER = get_logger(name='solver_client', file_prefix='core')


def main(args):
    if args.mode == 'server':
        mgr = SolverManager(args.endpoint_url)
        server = MCSServer(mgr, args.work_dir)
        server.run()
    else:
        if args.keyfile:
            eoaa, pkey = decode_keyfile(args.keyfile)
        else:
            eoaa, pkey = args.name, args.pkey
        mcs_console(Account(Ether(args.endpoint_url), eoaa, pkey), args.work_dir)


OPTIONS: List[Tuple[str, str, dict]] = [
    ('-m', '--mode', dict(
        action='store', required=True,
        choices=['server', 'client'])),
    ('-f', '--keyfile', dict(
        action='store', dest='keyfile',
        help='キーファイル')),
    ('-u', '--user', dict(
        action='store', dest='name',
        help='ログインユーザ(EOA Address)')),
    ('-k', '--privatekey', dict(
        action='store', dest='pkey',
        help='プライベートキー')),
    ('-e', '--endpoint', dict(
        action='store', dest='endpoint_url', required=True,
        help='Ethereum Provider Endpoint URL')),
    ('-w', '--workdir', dict(
        action='store', dest='work_dir', required=True,
        help='working dir where socket file is placed')),
]

if __name__ == '__main__':
    PARSER = argparse.ArgumentParser()
    for sname, lname, etc_opts in OPTIONS:
        PARSER.add_argument(sname, lname, **etc_opts)
    ARGS = PARSER.parse_args()
    main(ARGS)



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
import sys
from importlib import import_module
from typing import Any, Dict, Optional, Tuple, Type

from eth_typing import ChecksumAddress

from .bc.util import ADDRESS0
from .logger import get_logger
from .solver import BaseSolver

LOGGER = get_logger(name='plugin', file_prefix='core')
# the directory where plugins are placed
PLUGINS_PATH = os.getenv('PLUGINS_PATH', os.path.dirname(__file__) + '/../plugins')
SOLVER_CLASSNAME = 'Solver'


class PluginManager():

    def __new__(cls, *_args, **_kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        #                   filename: module
        self._modules: Dict[str, Any] = {}
        #                   operator_address:     (solver_class,     filename)
        self._solvers: Dict[ChecksumAddress, Tuple[Type[BaseSolver], str]] = {}

    def load(self, path: str = PLUGINS_PATH) -> None:
        if self._modules:
            LOGGER.debug('Plugin: already loaded')
            return

        # prepare for import_module
        sys.path.insert(0, path)

        for fname in os.listdir(path):
            filepath = '{}/{}'.format(path, fname)
            if not (os.path.isfile(filepath) and fname.endswith('.py')):
                continue
            try:
                mod_name = os.path.splitext(fname)[0]
                mod = import_module(mod_name)
                if hasattr(mod, SOLVER_CLASSNAME):
                    self._modules[fname] = mod
                    LOGGER.info('loaded %s from %s', SOLVER_CLASSNAME, fname)
            except Exception as err:
                LOGGER.exception(err)

        sys.path.remove(path)
        LOGGER.debug(self._modules)

    def set_default_solverclass(self, plugin_filename: str) -> None:
        self.set_solverclass(ADDRESS0, plugin_filename)

    def set_solverclass(self, operator_address: ChecksumAddress, plugin_filename: str) -> None:
        if plugin_filename not in self._modules.keys():
            raise Exception('No such plugin loaded: ' + plugin_filename)
        mod = self._modules[plugin_filename]
        if not hasattr(self._modules[plugin_filename], SOLVER_CLASSNAME):
            raise Exception('Not a solver: ' + plugin_filename)
        solver = getattr(self._modules[plugin_filename], SOLVER_CLASSNAME)
        self._solvers[operator_address] = (solver, plugin_filename)
        LOGGER.info(
            'Set solverclass %s.%s for operator %s',
            mod.__name__, solver.__name__, operator_address)

    def get_solverclass(self, operator_address: ChecksumAddress) -> Type[BaseSolver]:
        if operator_address in self._solvers.keys():
            return self._solvers[operator_address][0]
        if ADDRESS0 in self._solvers.keys():
            # return default if set.
            return self._solvers[ADDRESS0][0]
        return BaseSolver

    def get_plugin_filename(self, operator_address: ChecksumAddress) -> Optional[str]:
        if operator_address in self._solvers.keys():
            return self._solvers[operator_address][1]
        return None

    def is_pluginfile(self, filename: str) -> bool:
        return filename in self._modules.keys()



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from __future__ import annotations

import inspect
import json
import os
from time import sleep
from typing import ClassVar, Dict, Optional

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract as Web3Contract

from metemcyber.core.bc.account import Account
from metemcyber.core.logger import get_logger

LOGGER = get_logger(name='contract', file_prefix='core.bc')
MINIMAL_CONTRACT_ID = 'MetemcyberMinimal.sol:MetemcyberMinimal'
TX_RETRY_MAX = 10
TX_RETRY_DELAY_SEC = 2


def combined_json_path(contract_id: str) -> str:
    contract_src = contract_id.split(':')[0]
    src_dir = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(src_dir, 'contracts_data',
                            os.path.splitext(contract_src)[0] + '.combined.json')
    return filepath


def retryable_contract(cls):
    def __retryable(func):
        def wrapper(*args, **kwargs):
            for cnt in range(TX_RETRY_MAX):
                try:
                    return func(*args, **kwargs)
                except Exception as err:
                    LOGGER.exception(err)
                    sleep(TX_RETRY_DELAY_SEC)
                    LOGGER.warning(f'retrying:{cnt}')
                    continue
            raise Exception(f'Transaction retry count exceeds max({TX_RETRY_MAX}).')
        return wrapper

    # exclude methods defined in baseclass.
    excludes = [x[0] for x in inspect.getmembers(Contract)]
    for name, func in inspect.getmembers(cls):
        if name in excludes:
            continue
        if callable(getattr(cls, name)):
            setattr(cls, name, __retryable(func))
    return cls


class Contract():
    #                    contract_id: {address|placeholder: data}
    __deployed_libs: ClassVar[Dict[str, Dict[str, str]]] = {}
    __minimal_interface: ClassVar[Dict[str, str]] = {}
    #                              version: abi|bin: loaded data from combined json
    contract_interface: ClassVar[Dict[int, Dict[str, str]]] = {}  # overridden by sub class
    contract_id: ClassVar[Optional[str]] = None  # overridden by subclass
    _latest_version: ClassVar[int] = -1  # overridden by subclass

    def log_trace(self):
        try:
            cname = self.__class__.__name__
            frame = inspect.stack()[1][0]
            func = inspect.getframeinfo(frame).function
            args = {key: val for key, val
                    in inspect.getargvalues(frame).locals.items()
                    if key != 'self'}
            LOGGER.debug('%s(%s).%s%s', cname, self.address, func, args)
        finally:
            pass

    def log_success(self):
        try:
            cname = self.__class__.__name__
            frame = inspect.stack()[1][0]
            func = inspect.getframeinfo(frame).function
            LOGGER.debug('%s(%s).%s: succeeded', cname, self.address, func)
        finally:
            pass

    def __init__(self, account: Account):
        self.account: Account = account
        self._contract: Optional[Web3Contract] = None  # initialized by get()
        self.version: int = -1

    @property
    def web3(self) -> Web3:
        return self.account.web3

    @property
    def contract(self) -> Web3Contract:
        assert self._contract
        return self._contract

    @property
    def address(self):
        return self._contract.address if self._contract else None

    @classmethod
    def latest_version(cls):
        # currently, there's no way to figure out the latest version. it's ok, maybe.
        return cls._latest_version

    def new(self, *args, **kwargs):
        # pylint: disable=protected-access
        address = self.__class__.__deploy(self.account, *args, **kwargs)
        return self.get(address, id_check=False)

    def get(self, address: ChecksumAddress, id_check: bool = True):
        assert address
        if not Web3.isChecksumAddress(address):
            raise Exception('Invalid address: {}'.format(address))
        if not id_check:  # the case called from new()
            latest = self.__class__.latest_version()
            # pylint: disable=protected-access
            self.__class__.__load(latest)
            self._contract = self.web3.eth.contract(
                address=address, abi=self.__class__.contract_interface[latest]['abi'])
            return self

        minimal = self.web3.eth.contract(
            address=address, abi=self.__minimal_abi())
        try:
            tmp_id = minimal.functions.contractId().call()
        except Exception:
            tmp_id = 'unknown type of address'
        if tmp_id != self.__class__.contract_id:
            tmp_id = tmp_id.split(':', 1)[1] if ':' in tmp_id else tmp_id
            raise Exception(f'Invalid address. {address} is {tmp_id}.')
        self.version = minimal.functions.contractVersion().call()
        # pylint: disable=protected-access
        self.__class__.__load(self.version)
        self._contract = self.web3.eth.contract(
            address=address, abi=self.__class__.contract_interface[self.version]['abi'])
        return self

    @classmethod
    def register_library(cls, address, placeholder=''):
        #   for detail of placeholder, see below.
        #   https://solidity.readthedocs.io/en/latest/using-the-compiler.html
        assert address
        assert cls.contract_id
        if cls.contract_id in Contract.__deployed_libs.keys():
            raise Exception('already registered')
        if not placeholder:
            keccak = Web3.keccak(text=cls.contract_id).hex()[2:]  # cut 0x
            placeholder = '__$' + keccak[:34] + '$__'
        Contract.__deployed_libs[cls.contract_id] = {
            'address': address, 'placeholder': placeholder}
        LOGGER.debug(Contract.__deployed_libs)
        return placeholder

    @staticmethod
    def __minimal_abi() -> str:
        if not Contract.__minimal_interface:
            minimal_file = combined_json_path(MINIMAL_CONTRACT_ID)
            with open(minimal_file, 'r') as fin:
                meta_str = json.loads(fin.read())['contracts'][MINIMAL_CONTRACT_ID]['metadata']
            Contract.__minimal_interface['abi'] = json.loads(meta_str)['output']['abi']
            # omit bytecode
        return Contract.__minimal_interface['abi']

    @classmethod
    def __load(cls, version: int):
        if not cls.contract_id:
            raise Exception('contract_id is not defined: {}'.format(cls))
        if cls.contract_interface.get(version):
            return

        contract_interface = {}
        try:
            combined_file = combined_json_path(cls.contract_id) + (
                '' if version < 0 else f'-{version}')  # no suffix for latest
            with open(combined_file, 'r') as fin:
                combined_json = json.loads(fin.read())['contracts'][cls.contract_id]

            # Metadata (json nested in json) の追加
            contract_metadata = json.loads(combined_json['metadata'])
            contract_interface['abi'] = contract_metadata['output']['abi']

            # バイナリデータの追加
            bytecode = combined_json['bin']
            for lib in Contract.__deployed_libs.values():
                # Oops, link_code@solcx does not work well...
                # WORKAROUND: replace placeholder with address manually.
                bytecode = bytecode.replace(
                    lib['placeholder'], lib['address'][2:])  # cut 0x
            contract_interface['bin'] = bytecode

        except Exception as err:
            LOGGER.exception(err)
            raise Exception(
                f'Failed loading contract: {cls.contract_id} '
                f'version {"latest" if version < 0 else version}') from err
        cls.contract_interface[version] = contract_interface

    @classmethod
    def __deploy(cls, account: Account, *args, **kwargs):
        # コントラクトのチェーンへのデプロイ
        latest = cls.latest_version()
        if not cls.contract_interface.get(latest):
            cls.__load(latest)

        LOGGER.debug('deploying %s with args=%s, kwargs=%s',
                     cls.__name__, args, kwargs)

        # constructorに引数が必要な場合は指定
        if args or kwargs:
            func = account.web3.eth.contract(
                abi=cls.contract_interface[latest]['abi'],
                bytecode=cls.contract_interface[latest]['bin']).\
                constructor(*args, **kwargs)
        else:
            func = account.web3.eth.contract(
                abi=cls.contract_interface[latest]['abi'],
                bytecode=cls.contract_interface[latest]['bin']).\
                constructor()

        tx_hash = func.transact()
        tx_receipt = account.web3.eth.waitForTransactionReceipt(tx_hash)
        cls.gaslog('deploy', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError(
                'Contract deploy failed: {}'.format(cls.contract_id))

        LOGGER.info('deployed %s on address: %s',
                    cls.__name__, tx_receipt['contractAddress'])
        return tx_receipt['contractAddress']

    def event_filter(self, event_name, **kwargs):
        event = getattr(self.contract.events, event_name)
        return event.createFilter(**kwargs)

    @classmethod
    def gaslog(cls, func, tx_receipt):
        LOGGER.debug(
            '%s.%s: gasUsed=%d', cls.__name__, func, tx_receipt['gasUsed'])
